from algo.constant import Symbol
from algo.trading import OrderType, TradeResult
from algo.data.common import PricePoint
from algo.trading.oanda import RestApi


class CurrentState:
    def __init__(self) -> None:
        self._in_market = False
        self._open_price: float = None
        self._order_type: OrderType = None
        self._sl_size: float
        self.is_sl_close: bool = False

    @property
    def in_market(self) -> bool:
        return self._in_market

    @property
    def order_type(self) -> OrderType:
        return self._order_type

    def open(self, price_point: PricePoint, order_type: OrderType, sl_size=None):
        self._in_market = True
        self._order_type = order_type
        self._open_price = price_point.bid
        self._sl_size = sl_size
        self.is_sl_close = False

    def is_sl_time(self, price_point: PricePoint):
        """stop loss is negative .... why the hell not :) """
        return self.current_profit(price_point) < self._sl_size

    def current_profit(self, price_point: PricePoint) -> float:
        profit = self._open_price - price_point.bid
        if self._order_type != OrderType.SELL:
            profit *= -1
        return profit

    def close(self, price_point: PricePoint) -> TradeResult:
        """ close the trade and return profit"""
        self._in_market = False
        return TradeResult(
            self.current_profit(price_point),
            price_point.ask - price_point.bid
        )

    def __str__(self) -> str:
        return f'<State: order: {self.order_type}, open: {self._open_price}>'


class OandaTradingState(CurrentState):
    def __init__(self, default_symbol: Symbol) -> None:
        super().__init__()
        self._api = RestApi(default_symbol)
        # closing all trades for the case when prev run was interrupted
        self._api.close_all_trades()

    def open(self, price_point: PricePoint, order_type: OrderType, sl_size=None):
        self._api.open(order_type)
        return super().open(price_point, order_type, sl_size)

    def close(self, price_point: PricePoint) -> TradeResult:
        self._api.close_all_trades()
        return super().close(price_point)



from typing import Optional

from algo.strategy import STRATEGIES_REGISTER, StrategyDescriptor


def get_strategy_by_name(strategy_name: str) -> Optional[StrategyDescriptor]:
    """
    TBD: move to lib
    try to conside find ... not sure wich one is more performant
    """
    for strategy in STRATEGIES_REGISTER:
        if strategy.name == strategy_name:
            return strategy



import time
import json
import logging
from typing import List
import redis


from algo.data.common import PricePoint
from algo.config import config


log = logging.getLogger(__name__)


class RedisCache:
    def __init__(self) -> None:
        self._redis_connection = None

    def _init_redis_connection(self, force_reconnect=False):
        if force_reconnect or self._redis_connection is None:
            self._redis_connection: redis.Redis = redis.Redis.from_url(
                config.redis_url
            )

    def has_lock(self, cache_key: str):
        """TODO: implement lock with appropriate redis tools despite of workaround"""
        self._init_redis_connection()
        return self._redis_connection.get(cache_key) is not None

    def set_lock(self, cache_key: str):
        self._init_redis_connection()
        lock_is_set = False
        while not lock_is_set:
            try:
                self._redis_connection.set(cache_key, 1)
                lock_is_set = True
            except redis.ReadOnlyError as read_only_error:
                log.error(
                    'redis:ReadOnlyError', extra={'e': read_only_error}
                )
                time.sleep(1.0)
                self._init_redis_connection(True)

    def clear_lock(self, cache_key: str):
        self._init_redis_connection()
        return self._redis_connection.delete(cache_key)

    def has(self, cache_key: str):
        self._init_redis_connection()
        return self._redis_connection.get(cache_key) is not None

    def get(self, cache_key: str):
        """
        may be generator. Usage uf ujson may improve performance as this class essenstially does unjsoning
        """
        self._init_redis_connection()
        data = self._redis_connection.get(cache_key)
        all_data = []

        for data_portion in json.loads(data):
            all_data.append(PricePoint.from_json_ready(data_portion))

        return all_data

    def set(self, cache_key: str, data: List[PricePoint]):
        self._init_redis_connection()
        json_ready_data = []
        for data_portion in data:
            json_ready_data.append(data_portion.to_json_ready())
        serialized_data = json.dumps(json_ready_data)
        self._redis_connection.set(cache_key, serialized_data)



from enum import Enum
from typing import NamedTuple


class OrderType(Enum):
    BUY = 'BUY'
    SELL = 'SELL'


class TradeResult(NamedTuple):
    trade_profit: float
    trade_expenses: float

    @property
    def profit(self):
        return self.trade_profit - self.trade_expenses

    def __str__(self) -> str:
        return f'<TradeResult({self.profit}):\t{self.trade_profit}\t{self.trade_expenses})>'



import logging
from typing import List
import json

import boto3
from botocore.exceptions import ClientError

from algo.data.common import PricePoint, PriceStream


log = logging.getLogger(__name__)


TABLE_ALREADY_EXISTS_ERROR_CODE = 'ResourceInUseException'


def create_dynamo_cache_table(table_name, dynamodb):
    """
    # dynamodb: DynamoDBServiceResource

    ignoring billing arguments:
    ProvisionedThroughput
    and
    BillingMode
    for now setting billing mode to PAY_PER_REQUEST and will need to check cost of running and switch to
    Provision whenever it will make sense based on available load data
    """
    try:
        # TBD:  debug better on table creation it fails, on second run table already exists
        create_table_response = dynamodb.create_table(
            TableName=table_name,
            KeySchema=[
                {
                    'AttributeName': 'key',
                    'KeyType': 'HASH'
                }
            ],
            AttributeDefinitions=[
                {
                    'AttributeName': 'key',
                    'AttributeType': 'S'
                }
            ],
            BillingMode='PAY_PER_REQUEST'  # see comments
        )
        log.info('table-created-result', extra={'tr': create_table_response})
    except ClientError as exc:
        log.error(
            'error creating dynamo table',
            extra={
                'e': exc, 't': type(exc),
                'error-details': exc.response
            }
        )
        if exc.response['Error']['Code'] != TABLE_ALREADY_EXISTS_ERROR_CODE:
            raise


# 400K is actual max 300K is somewhat with safety range of 100K
MAX_CHAR_COUNT_PER_ENTRY = 300000


class DynamoCache():
    def __init__(self) -> None:
        self._client = boto3.client('dynamodb')
        self._res = boto3.resource('dynamodb')

        # consider moving to config ...
        self._dynamo_table_name = 'day-stream-data'  # prod value
        # self._dynamo_table_name = 'price-cache-day-experiement'
        self._table_exists = False
        self._dynamo_table = self._res.Table(self._dynamo_table_name)
        self._cache_data = {}

    def _verify_table_exists(self):
        if not self._table_exists:
            create_dynamo_cache_table(self._dynamo_table_name, self._res)
            self._table_exists = True

    def _has(self, cache_key: str):
        response = self._dynamo_table.get_item(
            Key={'key': cache_key}
        )
        return 'Item' in response

    def has(self, cache_key: str):
        """
        :PERF-IMPROVE-POTENTIAL:
        making repetitive read.... first reading to check, then reading to read if exists
        """
        self._verify_table_exists()
        chunked_key = format_chunked_key(cache_key, 1)
        return self._get_or_none(chunked_key) is not None

    def _get(self, cache_key: str) -> str:
        response = self._dynamo_table.get_item(
            Key={'key': cache_key}
        )
        return response['Item']['value']

    def _get_or_none(self, cache_key: str) -> str:
        response = self._dynamo_table.get_item(
            Key={'key': cache_key}
        )
        return response.get('Item', {'value': None})['value']

    def _set(self, cache_key: str, data: str):
        self._dynamo_table.put_item(
            Item={
                'key': cache_key,
                'value': data
            }
        )

    def has_lock(self, cache_key: str):
        self._verify_table_exists()
        return self._has(cache_key)

    def set_lock(self, cache_key: str):
        self._verify_table_exists()
        return self._set(cache_key, 1)

    def clear_lock(self, cache_key: str):
        self._verify_table_exists()
        self._dynamo_table.delete_item(
            Key={'key': cache_key}
        )

    def get(self, cache_key: str) -> PriceStream:
        """
        consider making iterative
        """
        self._verify_table_exists()
        chunk_number = 1
        formatted_cache_key = format_chunked_key(cache_key, chunk_number)
        data = self._get_or_none(formatted_cache_key)
        while data is not None:
            for data_portion in json.loads(data):
                yield PricePoint.from_json_ready(data_portion)
            chunk_number += 1
            formatted_cache_key = format_chunked_key(cache_key, chunk_number)
            data = self._get_or_none(formatted_cache_key)

    def has_meta(self, meta_key: str):
        self._verify_table_exists()
        return self._get_or_none(meta_key) is not None

    def set_meta(self, meta_key: str, data: any):
        self._verify_table_exists()
        serialized_data = json.dumps(data)
        self._set(meta_key, serialized_data)

    def set(self, cache_key: str, data: List[PricePoint]):
        """
        good oportunity to play OOP and decople
        chunk writer in a nice testable way :rofl:
        """
        self._verify_table_exists()
        data_chunk = JsonSizedArray()
        chunk_number = 1
        for data_portion in data:
            data_chunk.append(data_portion.to_json_ready())
            if data_chunk.size > MAX_CHAR_COUNT_PER_ENTRY:
                serialized_data = json.dumps(data_chunk.data)
                self._set(
                    format_chunked_key(cache_key, chunk_number),
                    serialized_data)
                data_chunk = JsonSizedArray()
                chunk_number += 1

        if len(data_chunk.data) > 0:
            serialized_data = json.dumps(data_chunk.data)
            self._set(format_chunked_key(
                cache_key, chunk_number), serialized_data
            )


def format_chunked_key(key, chunk_num: int):
    return f'{key}_{chunk_num}'


class JsonSizedArray():
    def __init__(self) -> None:
        self._data = []
        self._json_data_length_counter = 0

    def append(self, item: any):
        # 1 is for coma
        self._json_data_length_counter += len(json.dumps(item)) + 1
        self._data.append(item)

    @ property
    def size(self) -> int:
        # 2 for opening and closing braces
        return self._json_data_length_counter + 2

    @ property
    def data(self):
        return self._data



"""
this is pack of tools that should be used externally
"""
import logging
from datetime import datetime, timedelta, date
import symbol


from algo.data.reader import get_data_reader
from algo.data.common import Symbol
from algo.util import get_week_start_day
from algo.data.cache import has_week_data, set_day_cache


log = logging.getLogger(__name__)


def cache_from_influx(weeks_back: int = 19):
    """
    reads last 4 weeks and caches it into the DynamoCache.
    First week is current one if time of execution is at least Saturday UTC
    Otherwise start week will be previous.

    side note:
    data older than week of 2022-10-10 - 2022-10-15 is obtained from local(in house) listener
    """
    # deciding what week to start with,, current week is sunday to saturday so only day current week will be included
    # if today is Saturday. If today is sunday - the week will be previous, from the point of week of Sun - Sat.
    today = datetime.fromisoformat('2022-10-15')   # datetime.utcnow()
    start_week = get_week_start_day(today)
    if today.weekday() != 5:  # Saturday
        log.info('not a Saturday(UTC), so starting with prviouse week')
        start_week = get_week_start_day(today - timedelta(days=7))
    log.info('verifying/building if missing cache for last %s weeks, starting %s', weeks_back, start_week)
    data_reader = get_data_reader()

    for weeks_back in range(weeks_back):
        current_start_datetime = start_week - timedelta(days=weeks_back*7)
        current_start_date = current_start_datetime.date()

        log.info('check cache week started %s', current_start_date)
        for cur_symbol in Symbol:
            log.info('checking %s', cur_symbol)
            if has_week_data(cur_symbol, current_start_date):
                log.info('data is available for the %s, week started %s', cur_symbol, current_start_date)
                continue

            # otherwise making sure that there is enough data to cache it
            days_with_data = 0
            for day_of_week in range(7):
                current_week_date = current_start_date + timedelta(days=day_of_week)
                log.info('verifying dataset availability for the day %s', current_week_date)
                data_stream = data_reader.day_data(current_week_date, cur_symbol)
                all_data = list(data_stream)
                log.info('data count: %s', len(all_data))
                if len(all_data) > 0:
                    days_with_data += 1

            if days_with_data < 5:
                log.info('not enough(%s) data for caching %s, %s', days_with_data, cur_symbol, current_start_date)
                # to the next symbol
                continue

            log.info('days available: %s; caching data %s, %s', days_with_data, cur_symbol, current_start_date)

            for day_of_week in range(7):
                current_week_date = current_start_date + timedelta(days=day_of_week)
                data_stream = data_reader.day_data(current_week_date, cur_symbol)
                set_day_cache(cur_symbol, current_week_date, list(data_stream))



import logging
from typing import Generator

from algo.data.reader import PriceStream
# from algo.data.reporter import get_influx_strategy_output
from algo.indicator import MAverage
from algo.trading.state import CurrentState
from algo.trading import OrderType, TradeResult
from algo.data.common import PricePoint
from algo.util import d
from algo.strategy.common import BaseStrategyParams


class MAStrategyParams(BaseStrategyParams):
    first_avg_length: int
    second_avg_length: int


log = logging.getLogger(__name__)


def run_trading(data_stream: PriceStream, params: MAStrategyParams, state: CurrentState) -> Generator[TradeResult, None, None]:
    (mavg1_len, mavg2_len) = params
    mavg1 = MAverage(mavg1_len)
    mavg2 = MAverage(mavg2_len)
    history_price: PricePoint
    # state_reporter = get_influx_strategy_output()
    # trading_stream_id = 'influx-data-2022-10-31'
    for history_price in data_stream:
        # report_data = {}
        mavg1.tick(history_price.bid)
        mavg2.tick(history_price.bid)
        log.debug('price-data', extra=d(history_price, mavg1, mavg2))
        expected_direction = None  # kind of expected price direction

        if not (mavg1.ready() and mavg2.ready()):
            # skipping before ready
            continue
        # report_data.update({'mavg1': mavg1.value, 'mavg2': mavg2.value, 'price': history_price.bid})
        if mavg1.value > mavg2.value:
            expected_direction = OrderType.BUY
        else:
            expected_direction = OrderType.SELL

        if state.in_market and state.order_type != expected_direction:
            trade = state.close(history_price)
            log.info('closed-trade', extra=d(trade))
            # report_data.update({'close': history_price.bid})
            yield trade

        elif not state.in_market:
            log.info(
                'open-trade',
                extra=d(state, history_price, expected_direction)
            )
            # report_data.update({f'open_{expected_direction}': history_price.bid})
            state.open(history_price, expected_direction)

        # state_reporter.report_point(
        #     trading_stream_id,
        #     history_price.time,
        #     report_data
        # )



import logging
from datetime import datetime, timedelta, date
from dateutil import tz

from influxdb_client.client.flux_table import TableList


from algo.data.connection import DataConnection
from algo.data.common import PricePoint, PriceStream
from algo.util import d, dt_to_rfc3339, get_week_start_day, is_dt_naive
from algo.constant import Symbol
from algo.config import config


log = logging.getLogger(__name__)


class DataReader(DataConnection):
    def __init__(self, url, token=None, org=None):
        super().__init__(url, token, org)
        self._query_api = self._client.query_api()

    # TBD: consider removing
    def hour_data(self, hour_start: datetime, symbol: Symbol) -> PriceStream:
        """
        treats timezone naive data in hour_data as UTC
        """
        start = hour_start.replace(tzinfo=tz.tzutc()) if is_dt_naive(hour_start) else hour_start
        end = start + timedelta(hours=1)
        yield from self.get_data_stream(start, end, symbol)

    def day_data(self, day: date, symbol: Symbol) -> PriceStream:
        start = datetime.combine(
            day, datetime.min.time(), tz.tzutc()
        )
        end = start + timedelta(days=1)
        yield from self.get_data_stream(start, end, symbol)

    def week_data(self, week_day: date, symbol: Symbol) -> PriceStream:
        week_start_date = get_week_start_day(week_day)
        for days_offset in range(7):
            current_day = week_start_date + timedelta(days=days_offset)
            log.info('getting week data day by the day', extra=d(week_start=week_day, current_day=current_day))
            yield from self.day_data(current_day, symbol)

    def get_result_set(self, symbol: Symbol, start: datetime, end: datetime) -> TableList:
        query_str = f"""
            from(bucket: "tsdata")
              |> range(start: {dt_to_rfc3339(start)}, stop: {dt_to_rfc3339(end)})
              |> filter(fn: (r) => r["instrument"] == "{symbol.value}")
              |> filter(fn: (r) => r["_measurement"] == "price")
              |> pivot(rowKey: ["_time"], columnKey: ["_field"], valueColumn: "_value")
        """
        log.debug('q: %s', query_str)
        return self._query_api.query(query_str)

    def get_data_stream(
            self, start: datetime, end: datetime, symbol: Symbol
    ) -> PriceStream:
        log.debug('getting chunk starting', extra=d(start))
        results = self.get_result_set(symbol, start, end)
        log.info('query-result', extra={'r': results})

        for _time, ask, bid in results.to_values(columns=('_time', 'ask', 'bid')):
            yield PricePoint(_time, ask, bid)


def get_data_reader() -> DataReader:
    return DataReader(
        config.influx_url,
        config.influx_token,
        config.influx_org
    )



import logging
from datetime import datetime, timedelta
import time
from typing import List
from operator import itemgetter
import pytz

from influxdb_client.client.write_api import WriteType, WriteOptions
import boto3
from botocore.exceptions import ClientError
import ujson

from algo.config import config
from algo.data.connection import DataConnection
from algo.data.common import CompletePricePoint

log = logging.getLogger(__name__)

DATA_SENDING_INTERVAL = timedelta(milliseconds=5000)


class InfluxDataReporter(DataConnection):  # pylint: disable=too-few-public-methods
    def __init__(self, url: str, token: str, org: str):
        super().__init__(url, token, org)
        self._write_api = self._client.write_api(write_options=WriteOptions(WriteType.batching))
        self._last_sent_time = datetime.now()
        self._message_stack = []

    def report(self, price: CompletePricePoint):
        self._message_stack.append(price)
        if datetime.now() - self._last_sent_time > DATA_SENDING_INTERVAL:
            self._send_data()

    @staticmethod
    def _to_influx_data(price: CompletePricePoint):
        return {
            "measurement": 'price',
            "tags": {
                "instrument": price.instrument.value,
            },
            "time": price.time,
            "fields": {
                "ask": price.ask,
                "bid": price.bid,
            }
        }

    def _send_data(self):
        self._last_sent_time = datetime.now()
        start_time = time.time()
        messages_count = len(self._message_stack)
        json_body = [self._to_influx_data(d) for d in self._message_stack]
        self._message_stack.clear()
        self._write_api.write(bucket='tsdata', record=json_body)
        log.info('influx-data-sent', extra={
            'count': messages_count,
            'time_to_send_sec': time.time() - start_time
        })


class InfluxStrategyOutput(InfluxDataReporter):

    def report_point(self, data_stream: str, time_mark: datetime, data: dict):
        self._message_stack.append({
            "measurement": data_stream,
            "time": time_mark,
            "fields": data
        })
        if len(self._message_stack) > 1000:
            # self._write_api.write(
            #     bucket='tsdata',
            #     record=self._message_stack
            # )
            self._message_stack.clear()
            log.info('strategy-output 1k records sent', extra={
                'data_stream': data_stream
            })

    # this one has a space to lower current limit is 5 requests per second


ALREADY_EXISTS_ERROR_CODE = 'ResourceAlreadyExistsException'
INVALID_SEQUENCE_TOKEN_ERROR_CODE = 'InvalidSequenceTokenException'


class CloudWatchDataReporter:

    def _init_connection(self, log_group_name, log_stream_name):
        api = boto3.client('logs')

        try:
            api.create_log_group(logGroupName=log_group_name)
        except ClientError as exc:
            log.info(
                'error creating log group',
                extra={
                    'e': exc, 't': type(exc),
                    'error-details': exc.response
                }
            )
            if exc.response['Error']['Code'] != ALREADY_EXISTS_ERROR_CODE:
                raise

        try:
            api.create_log_stream(
                logGroupName=log_group_name,
                logStreamName=log_stream_name
            )
        except ClientError as exc:
            log.error(
                'error creating log stream',
                extra={
                    'e': exc, 't': type(exc),
                    'error-details': exc.response
                }
            )
            if exc.response['Error']['Code'] != ALREADY_EXISTS_ERROR_CODE:
                raise
        return api

    def __init__(self, log_group_name, log_stream_name) -> None:
        self._log_group_name = log_group_name
        self._log_stream_name = log_stream_name

        self._price_stack: List[CompletePricePoint] = []
        self._last_report_time = datetime.now()
        self._last_sequence_token = None
        self._api = self._init_connection(log_group_name, log_stream_name)

    def report(self, price: CompletePricePoint):
        self._price_stack.append(price)
        self._send_data()

    def _init_sequence_token(self):
        if self._last_sequence_token is None:
            streams = self._api.describe_log_streams(
                logGroupName=self._log_group_name, logStreamNamePrefix=self._log_stream_name
            )
            # uploadSequenceToken will be absent on first run for new stream
            self._last_sequence_token = streams['logStreams'][0].get(
                'uploadSequenceToken'
            )

    def _send_data(self):
        self._init_sequence_token()
        # TBD: need to catch potential errors and reconnect
        if datetime.now() - self._last_report_time > DATA_SENDING_INTERVAL:
            messages_count = len(self._price_stack)
            sending_start_time = time.time()
            log.info(
                'sending-cloudtrail-message',
                extra={
                    'dt': datetime.now() - self._last_report_time,
                    'data-count': messages_count
                }
            )
            self._last_report_time = datetime.now()
            records_batch = [{
                'timestamp': int(p.time.astimezone(pytz.utc).timestamp() * 1000),
                'message': ujson.dumps(p.dict)
            } for p in self._price_stack]

            log_events_args = {
                'logGroupName': self._log_group_name,
                'logStreamName': self._log_stream_name,
                'logEvents': sorted(records_batch, key=itemgetter('timestamp'), reverse=False),
            }
            self._price_stack.clear()

            if self._last_sequence_token is not None:
                log_events_args['sequenceToken'] = self._last_sequence_token

            self._send_data_failsafe(**log_events_args)
            log.info('cloud-trail-message-sent', extra={
                'count': messages_count,
                'time_to_send_sec': time.time() - sending_start_time
            })

    def _send_data_failsafe(self, **log_events_args):
        keep_trying = True
        while keep_trying:
            try:
                resp = self._api.put_log_events(**log_events_args)
                self._last_sequence_token = resp.get('nextSequenceToken', None)
                log.debug('put_log_event_response', extra={'response': resp})
                keep_trying = False
            except ClientError as exc:
                log.error(
                    'error sending data',
                    extra={
                        'e': exc, 't': type(exc),
                        'error-details': exc.response
                    }
                )
                # InvalidSequenceTokenException sometimes just happens(no good idea why)
                if exc.response['Error']['Code'] == INVALID_SEQUENCE_TOKEN_ERROR_CODE:
                    log.info('correct-sequence-token',
                             extra={
                                 'token': exc.response['expectedSequenceToken']}
                             )
                    log_events_args['sequenceToken'] = exc.response['expectedSequenceToken']
                    keep_trying = True
                else:
                    log.critical(
                        'unexpected error sending data to cloudwatch',
                        extra={
                            'e': exc, 't': type(exc),
                            'error-details': exc.response
                        })
                    raise


def get_influx_strategy_output() -> InfluxStrategyOutput:
    return InfluxStrategyOutput(
        'http://192.168.1.137:8086', None, 'some-org'
    )


def get_influx_data_reporter() -> InfluxDataReporter:
    return InfluxDataReporter(
        config.influx_url,
        config.influx_token,
        config.influx_org
    )


def get_cloud_watch_reporter() -> CloudWatchDataReporter:
    return CloudWatchDataReporter(
        config.aws_cloud_watch_log_group_name,
        config.aws_cloud_watch_log_stream_name
    )



from typing import Callable
import logging
from datetime import date, timedelta

from algo.data.cache.dynamo import DynamoCache
from algo.data.common import PriceStream, Symbol
from algo.util import get_week_start_day

from .dynamo import DynamoCache

CachableDataSource = Callable[[str, Symbol], PriceStream]

log = logging.getLogger(__name__)

dynamo_cache = DynamoCache()


def get_cached_week_data_stream(symbol: Symbol, week_day: date) -> PriceStream:
    week_start_date = get_week_start_day(week_day)
    for day_offset in range(7):
        current_day = week_start_date + timedelta(days=day_offset)

        data_generator = dynamo_cache.get(
            day_data_cache_key(symbol.value, current_day.isoformat())
        )
        yield from data_generator


def has_week_data(symbol: Symbol, day: date) -> bool:
    days_available = 0
    for day_offset in range(7):
        if has_day_data(symbol, day + timedelta(days=day_offset)):
            days_available += 1
    log.info('has_week_data: %s days available', days_available)
    return days_available >= 5


def set_day_cache(symbol: Symbol, day: date, data: list):
    cache_key = day_data_cache_key(symbol.value, day.isoformat())
    meta_key_name = meta_key(cache_key)
    log.info('setting cache %s', cache_key)
    log.info('setting cache meta %s', meta_key_name)
    dynamo_cache.set(cache_key, data)
    dynamo_cache.set_meta(meta_key_name, {'data_counter': len(data)})


def has_day_data(symbol: Symbol, day: date) -> bool:
    """
    externally used to check the availability of the cache
    """

    meta_key_for_day = meta_key(
        day_data_cache_key(
            symbol.value, day.isoformat()
        )
    )
    log.info('checking cache by meta %s', meta_key_for_day)
    return dynamo_cache.has_meta(meta_key_for_day)


def day_data_cache_key(symbol_str: str, date_str: str):
    return f'day-data-{symbol_str}-{date_str}'


def lock_key(cache_key: str) -> str:
    """lock key format"""
    return f'{cache_key}-lock'


def meta_key(cache_key: str) -> str:
    """meta key format"""
    return f'{cache_key}-meta'



#!/usr/bin/python3

# import json
from datetime import datetime, timedelta
import logging
import logging.config

from algo.data.common import CompletePriceStream, Symbol
# from algo.config import config
# from algo.constant import Symbol
# from algo.data.reader import PriceStream, get_data_reader
# from algo.strategy.moving_averages import MAStrategyParams, run_simulation
# from algo.trading import PricePoint
# from algo.util import d, now
# from algo.strategy import moving_average_sample
from algo.data.live import get_live_price_stream, get_complete_prices_generator
from algo.data.reader import get_data_reader
from algo.public import get_strategy_by_name
from algo.simulation import simulate_trading, simulate_trading_influx_data
from algo.util import d
from algo.config import config
from algo.data.reporter import InfluxDataReporter, InfluxStrategyOutput, get_cloud_watch_reporter
from algo.scripts import cache_from_influx

logging.config.fileConfig(
    './logging.conf',
    disable_existing_loggers=False
)

log = logging.getLogger(__name__)


# def serialize_data_record(data_record):
#     """
#     input is a FluxRecord
#     """
#     return data_record['result']['_measurement']


# def read_all(data: PriceStream):
#     data_points_counter = 0
#     total_length = 0  # in chars
#     total_ask = 0.0
#     total_bid = 0.0
#     for data_point in data:
#         # log.info('data-point', extra=d(data_point))
#         # log.info('data-point', extra=d(data_point[0].isoformat()))
#         total_ask += data_point[1]
#         total_bid += data_point[2]

#         data_serialized = data_point.to_json()
#         total_length += len(data_serialized)
#         data_points_counter += 1
#     return data_points_counter, total_length, total_ask, total_bid


def try_listener():
    """write the fucking tests"""
    config.set_config({
    })
    price_stream: CompletePriceStream = get_live_price_stream(
        price_reader=get_complete_prices_generator)
    # influx_reporter = get_influx_data_reporter()
    cw_reporter = get_cloud_watch_reporter()
    for price in price_stream:
        log.debug('price-point', extra=d(p=price))
        # influx_reporter.report(price)
        cw_reporter.report(price)


def verify_cache_works():
    data_reader = get_data_reader()
    log.info('loading cached data')
    data_iterator = data_reader.day_data(
        '2022-09-01 00:00:00+00:00', Symbol.JPY
    )
    all_data = list(data_iterator)
    log.info('done')
    log.info('loading influx no cache data')
    no_cahe_iterator = data_reader.day_data_no_cache(
        '2022-09-01 00:00:00+00:00', Symbol.JPY
    )
    data_no_cache = list(no_cahe_iterator)
    log.info('done')
    assert all_data == data_no_cache
    # data_iterator = data_reader.week_data(
    #     '2022-09-04 00:00:00+00:00', Symbol.JPY
    # )

    # data_counter = 0
    # for d in data_iterator:
    #     data_counter += 1
    #     if data_counter % 100 == 0:
    #         log.info('data-counter', extra={'counter': data_counter})


def cache_data():
    data_reader = get_data_reader()
    log.info('loading cached data')
    for s in Symbol:
        log.info('loading data', extra={'s': s})
        data_iterator = data_reader.week_data(
            '2022-10-20 00:00:00+00:00', s
        )
        all_data = list(data_iterator)
        log.info(
            'got-data',
            extra={
                'len': len(all_data),
                'symbol': s
            }
        )


def try_dynamo_cache():
    from algo.data.cache.dynamo import DynamoCache  # pylint: disable=all
    dc = DynamoCache()
    dc.set('test-data', {'d': 12})
    assert {'d': 12} == dc.get('test-data')
    dc.set_lock('sv')
    assert dc.has_lock('sv')
    assert not dc.has_lock('asdf')
    dc.clear_lock('sv')
    assert not dc.has_lock('sv')


def cache_data_from_local_server():
    dr = get_data_reader()
    current_week = datetime.fromisoformat('2022-09-22 00:00:00+00:00')
    end_week = datetime.fromisoformat('2022-10-20 00:00:00+00:00')
    while current_week < end_week:
        log.info('requesting data for the week', extra={
            'week-of': current_week
        })
        for s in Symbol:
            log.info('loading data', extra={'s': s})
            data_iterator = dr.week_data(
                current_week.isoformat(), s
            )
            all_data = list(data_iterator)
            log.info(
                'got-data',
                extra={
                    'len': len(all_data),
                    'symbol': s,
                    'week': current_week
                }
            )

        current_week += timedelta(days=7)


def playgraund():
    log.info('starting simulation')
    strategy = get_strategy_by_name('sample-moving-average')

    if strategy is not None:
        # '2022-11-09':  "simulation-result": [4.826000000000022, 3.057999999999936, 130]
        # '2022-10-10':  "simulation-result": [0.025999999999811507, 1.3840000000000146, 101]
        # simulate_trading(
        #     '2022-10-10', Symbol.JPY,
        #     strategy.module,
        #     strategy.args(4000, 5400)
        # )
        #  '2022-11-09': "simulation-result": [4.826000000000022, 3.057999999999936, 130]
        #  '2022-10-10': "simulation-result": [0.025999999999811507, 1.3840000000000146, 101]
        simulate_trading_influx_data(
            '2022-10-10', Symbol.JPY,
            strategy.module,
            strategy.args(4000, 5400)
        )


if __name__ == '__main__':
    log.info('getting started')
    # playgraund()
    # following is a line for a weekly  data move to  dynamoDB
    cache_from_influx()
    # end of a weekly  data move to  dynamoDB



class MemoryCache():
    def __init__(self) -> None:
        self._cache_data = {}

    def has(self, cache_key: str):
        return cache_key in self._cache_data

    def get(self, cache_key: str):
        return self._cache_data[cache_key]

    def set(self, cache_key: str, data):
        self._cache_data[cache_key] = data

    def has_lock(self, cache_key: str):
        return self.has(cache_key)

    def set_lock(self, cache_key: str):
        return self.set(cache_key, 1)

    def clear_lock(self, cache_key: str):
        del self._cache_data[cache_key]



import time
import logging
from typing import Callable, Union

from ujson import JSONDecodeError as UJSONDecodeError

import v20
from v20.errors import V20Timeout, V20ConnectionError
from v20.pricing import ClientPrice, PricingHeartbeat


from algo.config import config
from algo.data.common import PricePoint, PriceStream, CompletePricePoint, CompletePriceStream
from algo.constant import Symbol
from algo.util import d
from algo.metric import incr_message_counter


log = logging.getLogger(__name__)


def price_to_dict(price: ClientPrice):
    return {
        'type': type(price).__name__,
        'dict': price and price.dict()
    }


def heartbeat_to_dict(heartbeat: PricingHeartbeat):
    return {
        'type': type(heartbeat).__name__,
        'dict': heartbeat and heartbeat.dict()
    }


def get_prices_generator(stream) -> PriceStream:
    for msg_type, msg in stream.parts():
        log.debug(
            'got message', extra={
                'message_type': msg_type,
                'dict': msg.dict()
            }
        )
        incr_message_counter()
        if msg_type == 'pricing.PricingHeartbeat':
            log.debug('pricing.PricingHeartbeat', extra=heartbeat_to_dict(msg))
        elif msg_type == 'pricing.ClientPrice':
            log.debug('pricing.ClientPrice', extra=price_to_dict(msg))
            if PricePoint.is_valid_price_data(msg):
                yield PricePoint.from_oanda_price(msg)
        else:
            log.info('unexpected-stream-message', extra=d(msg))


def get_complete_prices_generator(stream) -> CompletePriceStream:
    for msg_type, msg in stream.parts():
        log.debug(
            'got message', extra={
                'message_type': msg_type,
                'dict': msg.dict()
            }
        )
        incr_message_counter()
        if msg_type == 'pricing.PricingHeartbeat':
            log.debug('pricing.PricingHeartbeat', extra=heartbeat_to_dict(msg))
        elif msg_type == 'pricing.ClientPrice':
            log.debug('pricing.ClientPrice', extra=price_to_dict(msg))
            if CompletePricePoint.is_valid_price_data(msg):
                yield CompletePricePoint.from_oanda_price(msg)
        else:
            log.info('unexpected-stream-message', extra=d(msg))


AnyPriceStream = Union[PriceStream, CompletePriceStream]
PriceReader = Callable[
    [any],  # I don't want to define Oanda's class that has method parts that returns generator of PricingHeartbeat | ClientPrice
    AnyPriceStream
]


def get_live_price_stream(symbol: Symbol = None, price_reader: PriceReader = get_prices_generator) -> AnyPriceStream:
    """
    first in a queue to cover with tests
    """
    symbol_str = None
    if symbol is not None:
        symbol_str = symbol.value
    else:
        symbol_str = Symbol.all_values()

    api = v20.Context(
        hostname=config.oanda_stream_host_name,
        token=config.oanda_token,
    )
    while True:
        # exit point is unexpected  error
        try:
            log.info('getting-new-stream')
            stream = api.pricing.stream(
                accountID=config.oanda_account_id,
                instruments=symbol_str,
                snapshot=True
            )

            yield from price_reader(stream)
        except (V20ConnectionError, V20Timeout) as err:
            # stream can timeout, so re-connecting
            log.error(
                'catched-v20error',
                extra=d(err, err_type=type(err)),
                exc_info=err
            )
            time.sleep(1.0)
        except UJSONDecodeError as json_decode_error:
            log.error('json-decode-error', extra=d(json_decode_error))

        except Exception as unexpected_error:
            # die bug log why exactly
            log.error(
                'unexpected-error',
                extra=(d(unexpected_error,
                         err_type=type(unexpected_error))),
                exc_info=unexpected_error
            )
            raise



import configparser

import setuptools

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

# following block reads deps from Pipfile nearby
# used as temporary fix to the unperfection of the world :) https://github.com/pypa/pipenv/issues/1263

install_requires = []
config = configparser.ConfigParser()
config.read('Pipfile')

if 'packages' not in config:
    raise Exception(
        'Unable to complete install_requires list. '
        'No packages section in Pipfile nearby.'
    )

for package in config['packages']:
    install_requires.append(package)


setuptools.setup(
    name='trade-utils',
    version='1.0.24',
    author='darnes',
    author_email='darnesmeister@gmail.com',
    license='MIT',
    description=(
        'Algo trading utils package. '
        'Zero test-coverage so please cross your fingers prior use.'
    ),
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/darnes/algo',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.9',
    include_package_data=True,
    install_requires=install_requires
)



import logging
from typing import Callable, Generator, NamedTuple

from algo.data.reader import PriceStream
from algo.trading import TradeResult
from algo.trading.state import CurrentState

log = logging.getLogger(__name__)
# for now money are floating... hopefully they will start soaring soon :rofl:
Money = float


class TradingResult(NamedTuple):
    profit: Money
    expenses: Money
    number_of_trades: int


BaseStrategyParams = NamedTuple


Strategy = Callable[
    [PriceStream, BaseStrategyParams, CurrentState],
    Generator[TradeResult, None, None]
]

Plan = Callable[
    [CurrentState],
    Generator[TradeResult, None, None]
]


class StrategyDescriptor(NamedTuple):
    name: str
    module: Strategy
    args: BaseStrategyParams
    description: str



import unittest
from datetime import date

from algo.util import get_week_start_day


class Test_get_week_start_day(unittest.TestCase):  # pylint: disable=invalid-name

    def test_week_start_day(self):

        expected_week_start = '2022-11-06'  # Sunday

        start_week_day = get_week_start_day(date.fromisoformat('2022-11-12'))  # Saturday
        self.assertEqual(start_week_day.isoformat(), expected_week_start)

        start_week_day = get_week_start_day(date.fromisoformat('2022-11-11'))  # Friday
        self.assertEqual(start_week_day.isoformat(), expected_week_start)

        start_week_day = get_week_start_day(date.fromisoformat('2022-11-10'))  # Thursday
        self.assertEqual(start_week_day.isoformat(), expected_week_start)

        start_week_day = get_week_start_day(date.fromisoformat('2022-11-09'))  # Wednesday
        self.assertEqual(start_week_day.isoformat(), expected_week_start)

        start_week_day = get_week_start_day(date.fromisoformat('2022-11-07'))  # Monday
        self.assertEqual(start_week_day.isoformat(), expected_week_start)

        start_week_day = get_week_start_day(date.fromisoformat(expected_week_start))  # Sunday iteself
        self.assertEqual(start_week_day.isoformat(), expected_week_start)



import logging

from influxdb_client import InfluxDBClient

log = logging.getLogger(__name__)


class DataConnection:  # pylint: disable=too-few-public-methods
    def __init__(self, url, token, org):
        self._client = InfluxDBClient(
            url=url, token=token, org=org
        )
        log.info('connected to influx', extra={'url': url})
